"""Team Fortress 2 Console Monitor."""

import argparse
import csv
import curses
import re
import threading
from pathlib import Path
from pprint import pformat

import libcurses
from loguru import logger

from tf2mon.admin import Admin
from tf2mon.conlog import Conlog
from tf2mon.fkey import FKeyManager
from tf2mon.gameplay import Gameplay
from tf2mon.hacker import HackerManager
from tf2mon.msgqueue import MsgQueueManager
from tf2mon.role import Role
from tf2mon.spammer import Spammer
from tf2mon.steamweb import SteamWebAPI
from tf2mon.ui import UI
from tf2mon.user import Team
from tf2mon.usermanager import UserManager


class TF2Monitor:
    """Team Fortress 2 Console Monitor."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, options: argparse.Namespace, config: {}) -> None:
        """Initialize monitor."""

        self.options = options
        self.config = config

        #
        self.steam_web_api = SteamWebAPI(
            dbpath=self.options.players,
            webapi_key=self.config["webapi_key"],
        )

        # send data to tf2 by writing to an `exec` script
        self.tf2_cfg_dir = Path(self.options.tf2_install_dir, "cfg")
        self.tf2_scripts_dir = Path(self.tf2_cfg_dir, "user")
        self.worker_path = Path(self.tf2_scripts_dir, "tf2-monitor-work.cfg")
        self.msgqueues = MsgQueueManager(self, self.worker_path)

        # prepare to receive data from tf2 by reading its console logfile
        self.conlog = Conlog(self)

        # inject any commands from the command line into the conlog
        self.conlog.inject_cmd_list(self.options.inject_cmds)

        # inject any commands from a file into the conlog
        if self.options.inject_file:
            logger.info(f"Reading `{self.options.inject_file}`")
            with open(self.options.inject_file, encoding="utf-8") as file:
                self.conlog.inject_cmd_list(file)

        # this application's admin console
        self.admin = Admin(self)

        # other options that inject commands...
        if self.options.breakpoint is not None:
            self.admin.set_single_step_lineno(self.options.breakpoint)

        #
        self.hackers = HackerManager(
            base=self.options.hackers_base,
            local=self.options.hackers_local,
        )

        #
        self.roles = Role.get_roles_by_name()
        self.sniper_role = self.roles["sniper"]

        #
        path = Path(__file__).parent / "data" / "weapons.csv"
        logger.info(f"Reading `{path}`")
        with open(path, encoding="utf-8") as _f:
            self.weapons = {weapon: self.roles[role] for role, weapon in csv.reader(_f)}

        #
        path = Path(__file__).parent / "data" / "racist.txt"
        logger.info(f"Reading `{path}`")
        lines = path.read_text(encoding="utf-8").splitlines()
        self._racist_re = (
            re.compile("|".join(lines), flags=re.IGNORECASE) if len(lines) > 0 else None
        )

        # users
        self.users = None

        # message queues
        self.kicks = self.msgqueues.addq("kicks")
        self.spams = self.msgqueues.addq("spams")

        #
        self.spammer = Spammer(self)

        # function keys
        self.fkeys = FKeyManager(self)
        self.fkeys.create_tf2_exec_script(self.tf2_scripts_dir / "tf2-monitor-fkeys.cfg")

        # admin command handlers
        self.regex_list = self.admin.regex_list

        # gameplay handlers
        self.gameplay = Gameplay(self)
        self.regex_list += self.gameplay.regex_list

        # function key handlers
        self.regex_list += self.fkeys.get_regex_list()

        #
        self.me = self.my = None
        self.ui = None
        self.reset_game()

    @property
    def appname(self) -> str:
        """Application name."""
        return self.config["app-name"]

    @property
    def cmd_prefix(self) -> str:
        """Return command prefix string."""
        # We bind TF2 function keys to a) perform an action [optional], and
        # b) `echo` a command message to the con_logfile to inform us the
        # function key was pressed. Tag all command messages to avoid
        # colliding with lines generated by TF2.
        return self.appname + "-"

    def run(self):
        """Run the Monitor."""
        libcurses.wrapper(self._run)

    def _run(self, win):
        # Build user-interface
        self.ui = UI(self, win)
        self.reset_game()

        # no need for threads if exiting at end of conlog
        if not self.options.follow:
            self.gameplay.repl()
            return

        # Read from conlog, write to display.
        thread = threading.Thread(name="game", target=self.gameplay.repl, daemon=True)
        thread.start()

        # main thread reads from keyboard/mouse, and writes to display
        threading.current_thread().name = "main"
        self.admin.repl()

    def reset_game(self):
        """Start new game."""

        logger.success("RESET GAME")

        self.users = UserManager(self)
        self.me = self.my = self.users.find_username(self.config["player_name"])
        self.me.assign_team(Team.BLU)
        self.my.display_level = "user"

        self.msgqueues.clear()
        if self.ui:
            self.ui.clear_chats()

    def breakpoint(self):
        """Drop into python debugger."""

        if self.conlog.is_eof:  # don't do this when replaying logfile from start
            curses.reset_shell_mode()
            breakpoint()  # pylint: disable=forgotten-debug-statement
            curses.reset_prog_mode()

    def kick_my_last_killer(self, attr):
        """Kick the last user who killed the operator."""

        if self.my.last_killer:
            self.my.last_killer.kick(attr)
        else:
            logger.warning("no last killer")

    def is_racist_text(self, text):
        """Return True if this user appears to be racist."""
        return self._racist_re.search(text)

    def dump(self) -> None:
        """Dump stuff."""

        # logger.success(pformat(self.__dict__))
        logger.success(pformat(self.users.__dict__))
        # pylint: disable=protected-access
        for user in self.users._users_by_username.values():
            logger.success(pformat(user.__dict__))



"""Role (job, function, position) of `User`."""

from dataclasses import dataclass
from typing import Dict


@dataclass
class Role:
    """Role (job, function, position) of `User`."""

    name: str  # match tokens in `con_logfile` and `weapons.csv`
    display: str  # for scoreboard

    @staticmethod
    def get_roles_by_name() -> Dict:
        """Return `dict` of `Role`s indexed by name."""

        roles = {}
        for display, name in [
            ("C", "scout"),
            ("S", "soldier"),
            ("P", "pyro"),
            ("D", "demo"),
            ("E", "engineer"),
            ("H", "heavy"),
            ("M", "medic"),
            ("N", "sniper"),
            ("Y", "spy"),
        ]:
            roles[name] = Role(name, display)

        return roles



"""Duel.

A contest with deadly weapons arranged between two people in order to
settle a point of honor.
"""

from dataclasses import InitVar, dataclass

from tf2mon.user import User


@dataclass
class Duel:
    """A contest between players."""

    killer: User
    victim: User
    weapon: InitVar[str]
    crit: InitVar[bool]
    key: str = None

    def __post_init__(self, weapon, crit) -> None:
        """Create hashable key for this unique "weapon-state".

        Also used as the formatted display value (for this unique "weapon-state").
        """

        name = self.killer.role.name if self.killer.role else ""
        crit = "+crit" if crit else ""
        perk = "+" + self.killer.perk if self.killer.perk else ""

        self.key = " ".join(
            [
                f"{name:8}",  # 8=len("engineer")
                f"{crit:5}",  # 5=len("+crit")
                f"{weapon:26}",  # 26=len("tf_projectile_pipe_remote")
                perk,
            ]
        ).rstrip()



"""Command line interface."""

import os
import sys
from pathlib import Path
from typing import List, Optional

import steam.steamid
import xdg
from libcli import BaseCLI
from loguru import logger

from tf2mon.conlog import Conlog
from tf2mon.logger import configure_logger
from tf2mon.tf2monitor import TF2Monitor


class Tf2monCLI(BaseCLI):
    """Command line interface."""

    _cachedir = xdg.xdg_cache_home() / __package__

    config = {
        # name of config file.
        "config-file": Path.home().joinpath(".tf2mon.toml"),
        #
        # toml [section-name].
        "config-name": "tf2mon",
        #
        # location of game to monitor
        "tf2_install_dir": Path.home().joinpath(
            "SteamLibrary", "steamapps", "common", "Team Fortress 2", "tf"
        ),
        #
        # basename relative to `tf2_install_dir`.
        "con_logfile": Path("console.log"),
        #
        # application name.
        "app-name": "TF2MON",
        # databases.
        "players": _cachedir / "steamplayers.db",
        "hackers-base": _cachedir / "playerlist.milenko-list.json",
        "hackers-local": _cachedir / "playerlist.tf2mon-list.json",
        # this player.
        "player_name": "Bad Dad",
    }

    exclude_print_config = [
        "app-name",
    ]

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.parser = self.ArgumentParser(
            prog=__package__,
            description=self.dedent(
                """
    Team Fortress II (`TF2`) Console Monitor (`%(prog)s`) is an interactive
    terminal application that displays real-time game state and player
    statistics of an actively running TF2 game.

    `%(prog)s` recognizes and tracks `cathook-bots`, and provides both
    application and in-game key-bindings to quickly `CALLVOTE KICK`
    hackers, mark players as racist, suspect, etc.

    `%(prog)s` reads TF2's console logfile (`con_logfile`), creates action
    scripts in TF2's `cfg` directory, and binds keys for the gamer to press
    in-game to take those actions; such as issue `SAY` and `CALLVOTE KICK`
    commands.

    Other in-game key-bindings format `CHAT` messages that include the
    name, K/D Ratio and duel-score with the player's last victim or killer.

    By default, `%(prog)s` starts reading the console logfile from its end
    (`--no-rewind`), and continues to `--follow` it until interrupted.
                """
            ),
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        rewind = self.parser.add_mutually_exclusive_group()
        arg = rewind.add_argument(
            "--rewind",
            dest="rewind",
            action="store_true",
            default=False,
            help="start from head of logfile",
        )
        self.add_default_to_help(arg)

        arg = rewind.add_argument(
            "--no-rewind",
            dest="rewind",
            action="store_false",
            help="start from tail of logfile",
        )
        self.add_default_to_help(arg)

        follow = self.parser.add_mutually_exclusive_group()
        arg = follow.add_argument(
            "--follow",
            dest="follow",
            action="store_true",
            default=True,
            help="follow end of logfile forever",
        )
        self.add_default_to_help(arg)

        arg = follow.add_argument(
            "--no-follow",
            dest="follow",
            action="store_false",
            help="exit at end of logfile",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--tf2-install-dir",
            metavar="DIR",
            default=Path(self.config["tf2_install_dir"]),
            type=Path,
            help="tf2 installation directory",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "con_logfile",
            default=Path(self.config["con_logfile"]),
            nargs="?",
            type=Path,
            help="tf2 console logfile",
        )
        self.add_default_to_help(arg)

        self.parser.add_argument(
            "--list-con-logfile",
            action="store_true",
            help="show tf2 console logfile path and exit",
        )

        self.parser.add_argument(
            "--trunc-con-logfile",
            action="store_true",
            help="truncate tf2 console logfile and exit",
        )

        self.parser.add_argument(
            "--clean-con-logfile",
            action="store_true",
            help="filter-out excluded lines from tf2 console logfile to stdout and exit",
        )

        group = self.parser.add_argument_group("Debugging options")

        group.add_argument(
            "--single-step",
            action="store_true",
            help="begin single-stepping at startup",
        )

        group.add_argument(
            "--break",
            dest="breakpoint",
            type=int,
            metavar="LINENO",
            help="single-step when logfile reaches `LINENO`",
        )

        group.add_argument(
            "--search",
            metavar="PATTERN",
            help="single-step logfile line matches `PATTERN`; add `/i` to ignore case.",
        )

        group.add_argument(
            "--inject-cmd",
            dest="inject_cmds",
            metavar="LINENO:CMD",
            action="append",
            help="inject `CMD` into logfile before `LINENO`",
        )

        group.add_argument(
            "--inject-file",
            metavar="FILE",
            help="read list of inject commands from `FILE`",
        )

        group = self.parser.add_argument_group("Database options")

        arg = group.add_argument(
            "--players",
            metavar="FILE",
            default=Path(self.config["players"]),
            type=Path,
            help="cache `steam` user data" "",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            "--hackers-base",
            metavar="FILE",
            default=Path(self.config["hackers-base"]),
            type=Path,
            help="upstream hackers database",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            "--hackers-local",
            metavar="FILE",
            default=Path(self.config["hackers-local"]),
            type=Path,
            help="local hackers database",
        )
        self.add_default_to_help(arg)

        group.add_argument(
            "--print-steamid",
            metavar="STEAMID",
            help="print `ISteamUser.GetPlayerSummaries` for `STEAMID` and exit" "",
        )

        self.parser.add_argument_group(
            "Configuration file",
            self.dedent(
                """
    The configuration file (see `--config FILE` below) defines local
    settings:

        [`%(prog)s`]

        tf2_install_dir = "/path/to/your/tf2/installation"
        webapi_key = "your-steamworks-webapi-key"
        player_name = "Your Name"
                """
            ),
        )

        self.parser.add_argument_group(
            "In-Game Controls, Numpad",
            self.dedent(
                """
    While playing TF2, gamer uses the `numpad` to perform actions, such as
    kicking a cheater, or taunting the last victim with a `chat` message
    customized with his name, k/d ratio and the duel-score with him.

    Messages are placed into queues, and may be popped off either end.

        Queues -->    Kicks     Admin     Spams
                        |         |         |
                        v         v         v
                   +-----------------------------+
        last/      |         |         |         |
        newest --> |   pop   |  pull   |   pop   |
                   |         |         |         |
                   |---------+---------+---------|
                   |         |         |         |
                   |  clear  |  clear  |  clear  |
                   |         |  both   |         |
                   |---------+---------+---------|
        first/     |         |         |         |
        oldest --> | popleft |  push   | popleft |
                   |         |         |         |
                   +-----------------------------+

    To vet any new players that have joined the game, gamer must press
    `NUMPAD-DOWNARROW` in-game. The monitor indicates when new players have
    arrived and that key should be pressed; or press it each time you die.

    When detected, `%(prog)s` pushes hackers onto the `Kicks` queue, and
    alerts the gamer, who may then press `NUMPAD-HOME` to issue `CHAT` and
    `CALLVOTE KICK` commands.

    When gamer kills an opponent, `%(prog)s` pushes a `taunt` onto the
    `Spams` queue; on death, a `gurgle`. Enable/disable with `F3`.

    The monitor can only push actions onto the queues; the gamer must pop
    them for the action to be taken, or clear the queue to discard.
                """
            ),
        )

        self.parser.add_argument_group(
            "Function Keys",
            self.dedent(
                """
    These function keys are available in-game and in the monitor:

        F1=HELP                 Display help.
        F2=TOGGLE-DEBUG         Control `say` vs `echo`.
        F3=TOGGLE-TAUNT         Enable taunts and gurgles.
        F4=TOGGLE-KD            Include kd-ratio in messages.
        F5=TOGGLE-USER-PANEL    Control contents of the user panel;
                                USER=show duels, weapons, captures, etc.
                                SPAMS=show SPAMS queue.
        F6=SWITCH-MY-TEAM       Join other team.
        F7=TOGGLE-SORT          Change scoreboard sort column.
        F8=TOGGLE-LOG-LOCATION  Various logger formats.
        KP_DEL=SINGLE-STEP      Start single-stepping.
        [=KICK-LAST-CHEATER     Kick last killer as cheater.
        ]=KICK-LAST-RACIST      Kick last killer as racist.
        \\=KICK-LAST-SUSPECT    Mark last killer as suspect.
                """
            ),
        )

        self.parser.add_argument_group(
            "Where to Operate",
            self.dedent(
                """
    `%(prog)s` works by reading the console logfile to which `TF2` logs
    messages during the game. `%(prog)s` can either `tail -f` an active
    game, or `--rewind` and replay saved logfiles.

        `Duel monitors`
            Run `%(prog)s` on a secondary monitor, while playing the game on
            the primary monitor. Best performance, but a pain to `Alt-Tab`
            between monitor and game.

        `Duel machines, ssh`
            ssh from another machine and run.

        `Duel machines, nfs`
            cross-mount TF2's directory to another box and run from there.
                    """
            ),
        )

        self.parser.add_argument_group(
            "Terminal Size",
            self.dedent(
                """
    `%(prog)s` requires a large terminal. Maximize the window, and use keys
    (maybe `Ctrl-Minus` and `Shift-Ctrl-Plus`) to resize. The wider the
    terminal, the more player data will be displayed:

        36x146 minimum
        42x173 display personaname
        52x211 display realname
        62x272 display age/location
                """
            ),
        )

        self.parser.add_argument_group(
            "Resizable Windows",
            self.dedent(
                """
    `Drag-and-drop` an interior border to resize the windows on either side.

    `Double-click` an interior border to enter `Resize Mode`.
        `scroll-wheel` and `arrows` move the border.
        `click`, `enter` or `esc` to exit.
                """
            ),
        )

        self.parser.add_argument_group(
            "Scoreboard",
            self.dedent(
                """
    `Single-click` user to highlight and follow.
    `Double-click` user to kick as cheater.
    `Triple-click` user to kick as racist.
                """
            ),
        )

    def main(self) -> None:
        """Command line interface entry point (method)."""

        configure_logger()

        if "webapi_key" not in self.config:
            self.parser.error("Missing config `webapi_key`")

        if self.options.con_logfile == self.config["con_logfile"]:
            # Not given on command line; prefix with effective parent.
            # (configured default is a basename relative to `tf2_install_dir`)
            self.options.con_logfile = Path(
                self.config["tf2_install_dir"],
                self.config["con_logfile"],
            )

        if self.options.single_step and not self.options.follow:
            self.parser.error("--no-follow not allowed with --single-step")

        if self.options.list_con_logfile:
            print(self.options.con_logfile)
            sys.exit(0)

        if self.options.trunc_con_logfile:
            Conlog(self.options.con_logfile).trunc()
            logger.warning(f"con_logfile {str(self.options.con_logfile)!r} truncated; Exiting.")
            sys.exit(0)

        if self.options.clean_con_logfile:
            for line in Conlog(self.options.con_logfile).filter_excludes():
                print(line)
            logger.warning(f"con_logfile {str(self.options.con_logfile)!r} cleaned; Exiting.")
            sys.exit(0)

        monitor = TF2Monitor(self.options, self.config)

        if self.options.print_steamid:
            monitor.steam_web_api.connect()
            steamid = steam.steamid.SteamID(self.options.print_steamid)
            if not steamid.is_valid():
                self.parser.error(f"invalid steamid `{self.options.print_steamid}`")
            sp = monitor.steam_web_api.find_steamid(steamid)  # noqa
            ic(sp)  # noqa
            sys.exit(0)

        if os.isatty(sys.stderr.fileno()):
            cmd = " ".join([self.parser.prog] + sys.argv[1:] + ["2>x"])
            logger.error(f"Please redirect `stderr`; try `{cmd}`")
            sys.exit(0)

        monitor.run()


def main(args: Optional[List[str]] = None) -> None:
    """Command line interface entry point (function)."""
    return Tf2monCLI(args).main()



"""Admin Console."""

import re
import threading

from loguru import logger

from tf2mon.hacker import HackerAttr
from tf2mon.regex import Regex


class Admin:
    """Admin Console."""

    def __init__(self, monitor):
        """Initialize Admin Console."""

        self.monitor = monitor
        self._single_step_event = threading.Event()
        self._is_single_stepping = self.monitor.options.single_step
        self._single_step_re = None

        if pattern := self.monitor.options.search:
            if pattern.startswith("/"):
                pattern = pattern[1:]
            self.set_single_step_pattern(pattern)

        if self._is_single_stepping:
            self.start_single_stepping()
        else:
            self._stop_single_stepping()
        self._single_step_at_eof = False

        # See fkeys.py for:
        #    tf2mon.fkeys.FKey(
        #        cmd='SINGLE-STEP',
        #        game_key='KP_DEL',
        #        curses_key=curses.KEY_DC,
        #        handler=lambda m: self.start_single_stepping(),
        #    ),

        # Commands operator may type into monitor admin console:
        self.regex_list = [
            # stop single-stepping
            Regex("^(c|cont|continue)$", lambda m: self._stop_single_stepping()),
            # stop single-stepping until eof, then single-step again
            Regex("^(r|run|g|go)$", lambda m: self.set_single_step_lineno(0)),
            # dump internals
            Regex("^dump$", lambda m: self.monitor.dump()),
            # pause when logfile reaches `lineno`.
            Regex(
                R"^(b|break|breakpoint)[= ](?P<lineno>\d+)$",
                lambda m: self.set_single_step_lineno(int(m.group("lineno"))),
            ),
            # pause when logfile reaches next line that matches `pattern`.
            Regex(
                "^/(?P<pattern>.*)$", lambda m: self.set_single_step_pattern(m.group("pattern"))
            ),
            # kick cheater
            Regex(
                R"^kick[= ](?P<userid>\d+)$",
                lambda m: self.monitor.users.kick_userid(
                    int(m.group("userid")), HackerAttr.CHEATER
                ),
            ),
            # kick racist
            Regex(
                R"^kkk[= ](?P<userid>\d+)$",
                lambda m: self.monitor.users.kick_userid(
                    int(m.group("userid")), HackerAttr.RACIST
                ),
            ),
            # mark suspect
            Regex(
                R"^suspect[= ](?P<userid>\d+)$",
                lambda m: self.monitor.users.kick_userid(
                    int(m.group("userid")), HackerAttr.SUSPECT
                ),
            ),
            # drop to python debugger
            Regex("^PDB$", lambda m: self.monitor.breakpoint()),
            # deprecated, legacy, support old conlogs.
            Regex(
                "^(TOGGLE-SCOREBOARD|TOGGLE-QUEUES|CHATS-POP|CHATS-POPLEFT|CHATS-CLEAR)$",
                lambda m: logger.debug("ignore"),
            ),
            # The following aren't commands, but other items that may be in
            # the conlog. They don't belong in `gameplay` because they're
            # not generated by tf2.
            # qvalve
            Regex("^(QVALVE) (.*)", lambda m: logger.log(m.group(1), m.group(2))),
        ]

    def start_single_stepping(self):
        """Begin prompting operator before processing each line from con_logfile."""

        self._is_single_stepping = True
        self._single_step_event.clear()

    def _stop_single_stepping(self):
        """End prompting operator before processing each line from con_logfile."""

        self._is_single_stepping = False
        self._single_step_event.set()

    def set_single_step_lineno(self, lineno=0):
        """Begin single-stepping at `lineno` if given else at eof."""

        if lineno:
            self.monitor.conlog.inject_cmd(lineno, "SINGLE-STEP")
        else:
            # stop single-stepping until eof, then single-step again
            self._stop_single_stepping()
            self._single_step_at_eof = True

    def set_single_step_pattern(self, pattern=None):
        """Begin single-stepping at next line that matches `pattern`.

        Pass `None` to clear the pattern.
        """

        if not pattern:
            self._single_step_re = None
            logger.log("ADMIN", "clear _single_step_re")
        else:
            flags = 0
            if pattern.endswith("/i"):
                pattern = pattern[:-2]
                flags = re.IGNORECASE
            elif pattern.endswith("/"):
                pattern = pattern[:-1]
            self._single_step_re = re.compile(pattern, flags)
            logger.log("ADMIN", f"set search={self._single_step_re}")

    def check_single_step(self, line):
        """Involve operator if `line` requires single-step attention.

        Called by the game thread (not the admin thread) for each line read
        from conlog, to wait for the operator (when necessary) before
        processing the line.

        If there's an active single-step pattern and last `line` read from
        con_logfile matches pattern, or if we are single-stepping, then
        wait for admin thread to release the lock.

        Args:
            line: last string read from con_logfile.

        Returns:
            immediately if gate is clear, else waits for it.
        """

        # start single stepping if pattern match
        if (
            self._single_step_re
            and not self._is_single_stepping
            and self._single_step_re.search(line)
        ):
            flags = "i" if (self._single_step_re.flags & re.IGNORECASE) else ""
            logger.log("ADMIN", f"break search /{self._single_step_re.pattern}/{flags}")
            self.start_single_stepping()

        level = "nextline" if self._is_single_stepping else "logline"
        logger.log(level, "-" * 80)
        logger.log(level, self.monitor.conlog.last_line)

        # check gate
        self._single_step_event.wait()
        if self._is_single_stepping:
            self._single_step_event.clear()

    def repl(self):
        """Admin console read-evaluate-process-loop."""

        while not self.monitor.conlog.is_eof or self.monitor.options.follow:

            self.monitor.ui.update_display()

            prompt = self.monitor.appname
            if self._is_single_stepping:
                prompt += " (single-stepping)"
            prompt += ": "

            if (line := self.monitor.ui.getline(prompt)) is None:
                logger.log("console", "quit eof")
                return

            if line == "":  # enter
                if self.monitor.conlog.is_eof:
                    logger.log("console", "<EOF>")
                else:
                    logger.trace("step...")
                self._single_step_event.set()
                continue

            logger.log("console", f"line={line!r}")

            cmd = line  # .upper()
            if "quit".find(cmd) == 0:
                logger.log("console", "quit")
                return

            if regex := Regex.search_list(cmd, self.regex_list):
                regex.handler(regex.re_match_obj)
            else:
                logger.error(f"bad admin command {cmd!r}")



"""A user of the game."""

import re
from collections import defaultdict
from enum import Enum

from fuzzywuzzy import fuzz
from loguru import logger

from tf2mon.hacker import HackerAttr


class Team(Enum):
    """Valid teams."""

    RED = 2
    BLU = 3


class UserState(Enum):
    """Valid states."""

    ACTIVE = "A"
    INACTIVE = "I"
    DELETE = "D"


class User:
    """A user of the game."""

    # pylint: disable=too-many-instance-attributes

    _milenko_re = re.compile(
        "|".join(
            [
                "I am a bot. I will leave in 15 seconds if there are no malicious"
                " bots to kick. https://milenko-tf2.github.io/",
                "No bots detected. GLHF! Anti-bot FAQ, Discord, and donations can"
                " be found at https://milenko-tf2.github.io/",
                "No bots detected. GLHF! https://incontestableness.github.io/milenko",
                "I am an anti-bot bot. I will leave in 15 seconds if there are no"
                " malicious bots to kick.",
                "Please press F2 to vote No, I leave automatically after about 15"
                " seconds if there are no malicious bots!",
                "No bots detected. Good luck, have fun!",
            ]
        )
    )

    def is_milenko_chat(self, chat):
        """Return True if this user appears to be milenko."""
        return self._milenko_re.search(chat.msg)

    _cheater_re = re.compile(
        "|".join(
            [
                "UNCLETOPIA.COM BEST MODERATED SERVERS! NO BOTS NO CHEATERS!",
            ]
        )
    )

    def is_cheater_chat(self, chat):
        """Return True first time chat appears to be from a cheater."""

        if self.cheater_chat_seen:
            return False
        self.cheater_chat_seen = self._cheater_re.search(chat.msg)
        return self.cheater_chat_seen

    def __init__(self, monitor, username):
        """Create `User`."""

        self.monitor = monitor
        self.username = username.replace(";", ".")

        if m := self.monitor._racist_re.search(self.username):
            self._clean_username = (
                m.string[: m.start()] + str("n" * (m.end() - m.start())) + m.string[m.end() :]
            )
        else:
            self._clean_username = self.username

        # too strict?
        if "\u0e31" in self.username:
            self.username = self.username.replace("\u0e31", "?")
            logger.warning(f"'\u0e31' in {self.username!r}")

        #
        self.username_upper = username.upper()
        self.userid = 0  # from status command
        self.steamid = None  # from status and tf_lobby_debug commands
        self.team = None
        self.ping = 0

        self.state = UserState.ACTIVE
        self.n_status_checks = 0
        self.nkills = 0
        self.nsnipes = 0
        self.role = None
        self.ndeaths = 0
        self.ncaptures = 0
        self.ndefenses = 0
        self.chats = []
        self.last_killer = None  # User
        self.last_victim = None  # User
        self.display_level = None
        self.selected = False
        self.perk = None

        # battles fought
        self.duels = []
        self.kills = defaultdict(lambda: defaultdict(int))
        self.deaths = defaultdict(lambda: defaultdict(int))
        self.opponents = defaultdict(User)

        # list of non-kill actions performed, like capture/defend.
        self.actions = []

        #
        self.hacker = None  # don't use unless vetted
        self.steamplayer = None  # don't use unless vetted
        self.vetted = False

        # When attempting to kick/track before hacker is available, this
        # indicates: a) the work (kick/track) has been postponed, and b)
        # the `HackerAttr.HackerAttr` to use when able to perform the work.

        self.work_attr = None

        # If this looks like a cheater we're tracking by name, mark him to
        # be kicked when his steamid becomes available. Doing this now to
        # notify the operator asap to `TF2MON-PUSH` steamids to us.
        # Careful, this might be a legitimate name-change, not a cheating name-stealer.

        self.cloner = None  # when this user is being cloned
        self.clonee = None  # when this user is the name-stealing clone

        if self._is_cheater_name(self.username):
            self.kick(HackerAttr.CHEATER)

        if self.monitor.is_racist_text(self.username):
            self.kick(HackerAttr.RACIST)

        self.cheater_chat_seen = False

    @property
    def kdratio(self):
        """Return kill/death ratio."""

        return self.nkills if not self.ndeaths else float(self.nkills / self.ndeaths)

    @property
    def points(self):
        """Return number of points scored."""

        return self.nkills + self.ncaptures + self.ndefenses

    @property
    def opposing_team(self):
        """Return opposing team."""

        return Team.RED if self.team == Team.BLU else Team.BLU

    @property
    def moniker(self):
        """Return name, optionally including his kill/death ratio."""

        if not self.monitor.ui.show_kd.value:
            return self._clean_username

        # pylint: disable=consider-using-f-string

        if self.ndeaths < 2:
            return "{!r} ({}/{})".format(self._clean_username, self.nkills, self.ndeaths)

        return "{!r} ({}/{}={:.1f})".format(
            self._clean_username, self.nkills, self.ndeaths, self.nkills / self.ndeaths
        )

    def duel_as_str(self, opponent, formatted=False):
        """Return string showing win/loss record against `opponent`."""

        nkills = len([x for x in self.duels if x.victim == opponent])
        ndeaths = len([x for x in self.duels if x.killer == opponent])
        return f"{nkills:2} and {ndeaths:2}" if formatted else f"{nkills} and {ndeaths}"

    def __repr__(self):

        team = f"{self.team.name}:" if self.team else ""
        return f"{team}{self.userid}={self.username!r}"

    def assign_teamno(self, teamno):
        """Assign this user to `teamno`."""

        try:
            team = Team(teamno)
        except ValueError as err:
            logger.error(f"{err} teamno {teamno!r}")

        self.assign_team(team)

    def assign_team(self, team):
        """Assign this user to `team`."""

        if isinstance(team, str):
            if team == Team.RED.name:
                team = Team.RED
            elif team == Team.BLU.name:
                team = Team.BLU
            else:
                logger.critical(f"bad team {team!r}")
                return

        if not self.team:
            logger.info(f"{self} joins {team}")

        elif self.team != team:
            logger.debug(f"{self} change from {self.team} to {team}")

        self.team = team

        # assign any unassigned opponents
        for duel in self.duels:
            if not duel.killer.team:
                duel.killer.assign_team(duel.victim.opposing_team)
            if not duel.victim.team:
                duel.victim.assign_team(duel.killer.opposing_team)

    def vet_player(self):
        """Examine user."""

        assert not self.vetted
        assert self.steamid

        self.steamplayer = self.monitor.steam_web_api.find_steamid(self.steamid)
        if self.steamplayer.is_legitimate_game_bot:
            self.steamplayer.personaname = self.username
            self.vetted = True
            self.work_attr = None
            return

        logger.debug(f"{self} SteamPlayer={self.steamplayer}")

        # has an existing user changed his name to this self.username?
        #  user := self._is_alias():
        #   vetted = "vetted" if user.vetted else "unvetted"
        #   logger.log("ALIAS", f"{self} steamid matches {vetted} {user}")
        #   # delete this alias from list
        #   self.state = UserState.DELETE
        #   # change name
        #   user.username = self.username
        #   user.username_upper = self.username.upper()
        #   # re-vet with new name
        #   user.vetted = False
        #   user.hacker = None
        #   user.vet_player()
        #   return

        # known hacker?
        self.hacker = self.monitor.hackers.lookup_steamid(self.steamid)
        if self.hacker:
            logger.log("hacker", self.hacker)
        else:
            logger.trace(f"{self} is not a known hacker")

        # should he be known?
        if self.work_attr and not self.hacker:
            # yes, work had been postponed until steamid now available
            self.hacker = self.monitor.hackers.add(self.steamid, self.work_attr, self.username)
            logger.log(self.work_attr.name, f"{self} created {self.hacker}")

        if self.hacker:
            # he's known
            self.hacker.track_appearance(self.username)
            self.monitor.hackers.save_database()

            self.display_level = self.hacker.attributes[0].name

            if self.hacker.is_banned:
                self.kick()
            elif self.hacker.is_suspect:
                logger.log(self.display_level, f"{self._clean_username!r} is here")

        # vet only once
        self.vetted = True
        self.work_attr = None

    def kick(self, attr=None):
        """Kick this user."""

        if self.userid in self.monitor.users.kicked_userids:
            logger.debug(f"already kicked userid {self.userid}")
            return

        if self._track(attr or HackerAttr.CHEATER):
            # work postponed until steamid becomes available
            return

        if self.hacker.is_milenko:
            # don't kick milenko
            msg = str(
                f"say {self.monitor.appname} Hello Milenko "
                f"{self.username!r} steamid {self.steamid!r}, thank you."
            )
            self.monitor.kicks.push(msg)
            return

        if self.hacker.is_suspect:
            # don't kick suspects
            logger.debug(f"not kicking suspect {self.username!r} steamid {self.steamid!r}")
            return

        # work

        msg = f"say {self.monitor.appname} ALERT: "
        if self.hacker.is_racist:
            msg += f"RACIST {self._clean_username!r}"
        elif self.clonee:
            msg += f"NAME-STEALER {self.username!r}"
        else:
            msg += f"CHEATER {self.username!r}"

        msg += " is here"
        cmd = f"CALLVOTE KICK {self.userid}"
        msg += f", {cmd}"

        self.monitor.kicks.push(msg)
        self.monitor.kicks.push(cmd)
        self.monitor.users.kicked_userids[self.userid] = True

    def _track(self, attr):

        self.display_level = attr.name

        if not self.steamid:
            if not self.work_attr:
                self.work_attr = attr
                logger.log(
                    self.display_level, f"{self} needs steamid, Press KP_DOWNARROW to PUSH"
                )
                self.monitor.ui.notify_operator = True
                self.monitor.ui.sound_alarm = True
            return True  # postpone work

        if not self.work_attr:
            # not postponed
            self.work_attr = attr
        # else un-postponed

        # work
        if self.hacker:
            if self.work_attr not in self.hacker.attributes:
                self.hacker.attributes.append(self.work_attr)
                logger.log(self.display_level, f"{self} added {self.work_attr} to {self.hacker}")
                self.monitor.hackers.save_database()
            else:
                logger.debug(f"{self} hacker {self.hacker} already {self.work_attr}")
        else:
            self.hacker = self.monitor.hackers.add(self.steamid, self.work_attr, self.username)
            logger.log(self.display_level, f"{self} created hacker {self.hacker}")
            self.monitor.hackers.save_database()

        # work performed
        self.work_attr = None
        return False

    _cheaters_re = re.compile(
        "|".join(
            [
                r"^(\(\d+\))?Sydney",
                r"Swonk Bot",
                r"spoooky braaaap",
                r"Bot Removal Service",
            ]
        )
    )

    def _is_cheater_name(self, name):

        if self._cheaters_re.search(name):
            return True

        for user in [x for x in self.monitor.users.active_users() if x.vetted and not x.hacker]:
            ratio = fuzz.ratio(name, user.username)
            if ratio > 80:
                logger.log("FUZZ", f"ratio {ratio} {name!r} vs {user.username!r}")
                # Careful, this might be a legitimate name-change, not a cheating name-stealer.
                user.cloner = self  # point the original user to the clone
                self.clonee = user  # point the clone to the original user
                return True

        return False

    def _is_alias(self):
        """Return the `User` that matches this user's steamid, else None."""

        users = [
            x
            for x in self.monitor.users.active_users()
            if x != self and x.steamid == self.steamid
        ]
        return users[0] if users else None



"""Collection of `User` objects."""

from loguru import logger

from tf2mon.steamplayer import SteamPlayer
from tf2mon.ui import SORT_ORDER
from tf2mon.user import User, UserState


class UserManager:
    """Collection of `User` objects."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, monitor):
        """Initialize `User` manager."""

        self.monitor = monitor

        # The pool.

        self._users_by_username = {}
        self._users_by_userid = {}
        self._users_by_steamid = {}
        self._teams_by_steamid = {}
        self.kicked_userids = {}

        # Some methods return a sorted list of users.
        # Names must match scoreboard column headings.

        self._sort_keys = {
            SORT_ORDER.K: lambda x: (-x.nkills, x.username_upper),
            SORT_ORDER.KD: lambda x: (-x.kdratio, -x.nkills, x.username_upper),
            SORT_ORDER.STEAMID: lambda x: (x.steamid.id if x.steamid else 0, x.username_upper),
            SORT_ORDER.USERNAME: lambda x: x.username_upper,
        }
        self._sort_key = self._sort_keys[SORT_ORDER.KD]

    def set_sort_order(self, sort_order):
        """Set sort-key used by `active_team_users`."""

        self._sort_key = self._sort_keys.get(sort_order)

    def delete(self, user):
        """Remove `user` from pool."""

        try:
            if user.last_killer and user.last_killer.last_victim == user:
                user.last_killer.last_victim = None

            if user.last_victim and user.last_victim.last_killer == user:
                user.last_victim.last_killer = None

            if user.cloner:
                user.cloner.clonee = None

            if user.clonee:
                user.clonee.cloner = None

            del self._users_by_username[user.username]
            del self._users_by_userid[user.userid]

            if user.steamid:
                del self._users_by_steamid[user.steamid]
                del self._teams_by_steamid[user.steamid]

        except KeyError:
            pass

    def find_username(self, username):
        """Return `User` with the matching `username` from pool.

        Create and add to pool, if not already there.
        User is always "found"; an object is always returned.
        """

        username = username.replace(";", ".")

        if not (user := self._users_by_username.get(username)):
            user = User(self.monitor, username)
            self._users_by_username[user.username] = user
            logger.log("ADDUSER", user)

        # reset inactivity counter
        user.n_status_checks = 0
        return user

    def status(self, userid, username, steamid, ping) -> None:
        """Respond to `gameplay.status` event."""

        user = None

        if steamid.id != SteamPlayer.BOT_STEAMID and (
            user := self._users_by_steamid.get(steamid)
        ):
            if user.username and user.username != username:
                logger.warning(f"{steamid} changed username from {user.username} to {username}")
            if user.userid and user.userid != userid:
                logger.warning(f"{steamid} changed userid from {user.userid} to {userid}")

        if not user:
            user = self.find_username(username)
            if user.userid and user.userid != userid:
                logger.warning(f"{username} changed userid from {user.userid} to {userid}")
            if user.steamid and user.steamid != steamid:
                logger.error(f"{username} changed steamid from {user.steamid} to {steamid}")

        if not user.userid:
            user.userid = userid
            self._users_by_userid[userid] = user

        if not user.steamid:
            user.steamid = steamid
            self._users_by_steamid[steamid] = user

        #
        user.ping = ping
        logger.log("STATUS", user)

        #
        if not user.team and (team := self._teams_by_steamid.get(steamid)):
            user.assign_team(team)

    def lobby(self, steamid, team):
        """Respond to `gameplay.lobby` event."""

        if old_team := self._teams_by_steamid.get(steamid):
            # if we've seen this steamid before...
            if old_team != team:
                # ...and
                logger.warning(f"{steamid} changed team from {old_team} to {team}")
        else:
            logger.log("ADDLOBBY", f"{team} {steamid}")

        #
        self._teams_by_steamid[steamid] = team

    def active_users(self):
        """Return list of active users."""

        yield from [x for x in self._users_by_username.values() if x.state == UserState.ACTIVE]

    def active_team_users(self, team):
        """Return list of active users on `team`."""

        assert self._sort_key

        yield from sorted(
            [
                x
                for x in self._users_by_username.values()
                if x.state == UserState.ACTIVE and x.team == team
            ],
            key=self._sort_key,
        )

    def kick_userid(self, userid, attr):
        """Kick `userid` reason `attr`."""

        if user := self._users_by_userid.get(userid):
            user.kick(attr)
        else:
            logger.error(f"bad userid {userid!r}")

    _max_status_checks = 2

    def check_status(self):
        """Delete users that appear to have left the game.

        We are called in response to TF2MON-PUSH; which may have came long
        before `status` finished sending everything; and we don't have a
        way to detect when `status` finishes; so we'll never be completely
        current. That's why `_max_status_checks` should be at least 2 or 3.
        """

        for user in list(self._users_by_username.values()):
            if user == self.monitor.me:
                continue
            user.n_status_checks += 1
            if user.n_status_checks == self._max_status_checks:
                logger.log(
                    "Inactive",
                    f"{user.n_status_checks} > {self._max_status_checks} {user.username!r}",
                )
                self.delete(user)

    def switch_teams(self):
        """Switch teams."""

        for user in self.active_users():
            user.assign_team(user.opposing_team)



import pytest

from tf2mon.cli import main


def test_version():
    with pytest.raises(SystemExit) as err:
        main(["--version"])
    assert err.value.code == 0


def test_help():
    with pytest.raises(SystemExit) as err:
        main(["--help"])
    assert err.value.code == 0


def test_md_help():
    with pytest.raises(SystemExit) as err:
        main(["--md-help"])
    assert err.value.code == 0


def test_long_help():
    with pytest.raises(SystemExit) as err:
        main(["--long-help"])
    assert err.value.code == 2


def test_bogus_option():
    with pytest.raises(SystemExit) as err:
        main(["--bogus-option"])
    assert err.value.code == 2


def test_print_config():
    with pytest.raises(SystemExit) as err:
        main(["--print-config"])
    assert err.value.code == 0


def test_print_url():
    with pytest.raises(SystemExit) as err:
        main(["--print-url"])
    assert err.value.code == 0


def test_debug():
    with pytest.raises(SystemExit) as err:
        main(["-X"])
    assert err.value.code == 0


def test_list_con_logfile():
    with pytest.raises(SystemExit) as err:
        main(["--list-con-logfile"])
    assert err.value.code == 0


@pytest.mark.parametrize(
    ("config_file", "expected"),
    [
        # (
        #     "/dev/null",
        #     str(
        #         "[tf2mon]\n"
        #         'config-file = "/dev/null"\n'
        #         'tf2_install_dir = "/home/russel/SteamLibrary/'
        #         'steamapps/common/Team Fortress 2/tf"\n'
        #         'con_logfile = "console.log"\n',
        #     ),
        # ),
        # ("tests/fixtures/file1.toml", None),
        # ("tests/fixtures/file2.toml", None),
        # ("tests/fixtures/file3.toml", None),
    ],
)
def test_config_files(config_file, expected):

    print()
    try:
        main(["-v", "--config", config_file, "--print-config"])
    except SystemExit as err:
        print(f"err: {err}")
        # assert err.value.code == 0
        print(f"Expected: {expected}")
    else:
        pytest.fail("didn't exit")



"""Interface to `ISteamUser.GetPlayerSummaries`."""

import sqlite3
import time

import steam.steamid
import steam.webapi
from loguru import logger

from tf2mon.steamplayer import SteamPlayer


class SteamWebAPI:
    """Interface to `ISteamUser.GetPlayerSummaries`.

    Results are cached to avoid banging the server.
    """

    def __init__(self, webapi_key, dbpath, max_age=2 * 60 * 60):
        """Initialize interface to Steam_Web_API.

        The interface cannot be used until `connect` is called.
        """

        #
        self._webapi_key = webapi_key
        self._webapi = steam.webapi.WebAPI(key=self._webapi_key)

        #
        self._dbpath = dbpath
        self._max_age = max_age
        self._con = None  # set in `connect`.
        self._cur = None  # set in `connect`.

        #
        self._nbots = 0  # incremented by `find_steamid`.

    def connect(self):
        """Prepare to use interface to Steam_Web_API.

        This step is separate from the constructor to allow connecting from
        a different thread.
        """

        logger.debug(f"connecting to sqlite3 at `{self._dbpath}`")
        self._con = sqlite3.connect(self._dbpath)
        self._con.row_factory = sqlite3.Row
        self._cur = self._con.cursor()

        self._cur.execute(
            """create table if not exists steamplayers(
                steamid integer primary key,
                personaname text,
                profileurl text,
                personastate text,
                realname text,
                timecreated integer,
                loccountrycode text,
                locstatecode text,
                loccityid text,
                mtime integer)"""
        )

        self._con.commit()

    def find_steamid(self, steamid):
        """Lookup and return `SteamPlayer` with matching `steamid`.

        Use web service to get "Player Summary" of given `steamid`.
        Create dummy object for game bots.

        Args:
            steamid: `steam.steamid.SteamID` of user to find.
        """

        if steamid.id == int(SteamPlayer.BOT_S_STEAMID):
            # create a dummy steamid for this bot; (not a hacker, a real game bot)
            self._nbots += 1
            return SteamPlayer(
                {
                    "steamid": steamid.id,
                    "personaname": None,
                    "profileurl": "",
                    "personastate": 0,
                    "realname": "",
                    "timecreated": int(time.time()) - (self._nbots * 86400),
                    "loccountrycode": "US",
                    "locstatecode": "IL",
                    "loccityid": "CHGO",
                }
            )

        # it's not a game bot; look in database.
        try:
            self._cur.execute("select * from steamplayers where steamid=?", (steamid.id,))
        except Exception as err:
            logger.critical(err)
            raise
        #
        if row := self._cur.fetchone():
            # convert tuple-like sqlite3.Row to json document
            steamplayer = SteamPlayer({k: row[k] for k in row.keys()})
            if steamplayer.mtime > int(time.time()) - self._max_age:
                # logger.debug('current')
                return steamplayer
            # logger.debug('expired')
        # else:
        #    logger.debug('notfound')

        # not current or not in database; ping server.
        player_summaries = self._get_player_summaries([steamid])

        if len(player_summaries) < 1:
            # unexpected!
            return SteamPlayer(
                {
                    "steamid": steamid.id,
                    "personaname": "???",
                    "profileurl": "",
                    "personastate": "?",
                    "realname": "",
                    "timecreated": int(time.time()),
                    "loccountrycode": "",
                    "locstatecode": "",
                    "loccityid": "",
                }
            )

        # construct.
        steamplayer = SteamPlayer(player_summaries[0])

        # update database.
        try:
            self._cur.execute(
                "replace into steamplayers values(?,?,?,?,?,?,?,?,?,?)",
                (
                    steamplayer.steamid.id,
                    steamplayer.personaname,
                    steamplayer.profileurl,
                    steamplayer.personastate,
                    steamplayer.realname,
                    steamplayer.timecreated,
                    steamplayer.loccountrycode,
                    steamplayer.locstatecode,
                    steamplayer.loccityid,
                    int(time.time()),
                ),
            )
        except Exception as err:
            logger.critical(err)
            raise
        self._con.commit()

        #
        return steamplayer

    def _get_player_summaries(self, steamids):

        jdoc = self._webapi.call(
            "ISteamUser.GetPlayerSummaries", steamids=",".join([str(x.as_64) for x in steamids])
        )

        return list(jdoc["response"]["players"])



"""User Interface module."""

import contextlib
import curses
import sys
import time
from enum import Enum

import libcurses
from loguru import logger

from tf2mon.scoreboard import Scoreboard
from tf2mon.toggle import Toggle
from tf2mon.user import Team, UserState

# from playsound import playsound


# These would have been defined within class UI (because they're only used internally)
# but `pydoc` doesn't display their values when defined there; it does when defined here.

USER_PANEL = Enum("_user_panel_enum", "USER SPAMS")
USER_PANEL.__doc__ = "What to display in the user window"

LOG_LOCATION = Enum("_log_location_enum", "MOD FILE THREAD NOLOC")
LOG_LOCATION.__doc__ = (
    "How to format the location field of messages displayed in the logger window."
)

SORT_ORDER = Enum("_so", "STEAMID K KD USERNAME")
SORT_ORDER.__doc__ = "Column to sort the scoreboard by." ""


class UI:
    """User Interface class."""

    # pylint: disable=too-many-instance-attributes

    _log_locations = {
        LOG_LOCATION.MOD: "{module}:{line}",
        LOG_LOCATION.FILE: "{file}:{line}:{function}",
        LOG_LOCATION.THREAD: "{thread.name}:{file}:{line}:{function}",
        LOG_LOCATION.NOLOC: None,
    }

    def __init__(self, monitor, win: curses.window):
        """Initialize User Interface instance.

        Args:
            monitor:    `TF2Monitor` instance.
            win:        curses window to use; likely `stdscr`.
        """

        self.monitor = monitor
        self.max_users = 32

        # create empty grid
        self.grid = libcurses.Grid(win)

        # promote some grid methods.
        self.redraw = self.grid.redraw
        self.refresh = self.grid.refresh

        # `self._build_grid` creates these windows, sized to fill `win`.
        self.chatwin_blu: curses.window = None
        self.chatwin_red: curses.window = None
        self.scorewin_blu: curses.window = None
        self.scorewin_red: curses.window = None
        self.user_win: curses.window = None
        self.logger_win: curses.window = None
        self.status_win: curses.window = None
        self.cmdline_win: curses.window = None

        # `register_builder` 1) calls `_build_grid` and 2) configures
        # `KEY_RESIZE` to call it again each time that event occurs.
        self.grid.register_builder(self._build_grid)

        #
        self.monitor.fkeys.register_curses_handlers()

        #
        self._curses_logwin = libcurses.LoggerWindow(self.logger_win)
        self._curses_logwin.set_verbose(self.monitor.options.verbose)
        self.colormap = self._curses_logwin.colormap

        #
        self._scoreboard = Scoreboard(
            self.monitor,
            self.scorewin_blu,
            self.colormap[Team.BLU.name],
            self.scorewin_red,
            self.colormap[Team.RED.name],
        )

        #
        self.debug_flag = Toggle("_df", [False, True])
        self.taunt_flag = Toggle("_tf", [False, True])
        self.gurgle_flag = Toggle("_gf", [True, False])

        # the USER window displays either the USER_PANEL.USER or the USER_PANEL.SPAMS
        self.user_panel = Toggle("_up", USER_PANEL)

        # options when displaying USER_PANEL.USER
        self.show_duels = Toggle("_sd", [True, False])
        self.show_actions = Toggle("_sa", [True, False])
        self.show_chats = Toggle("_sc", [False, True])

        # how to format the location field of messages displayed in the logger window.
        self.log_location = Toggle("_ll", LOG_LOCATION)
        self.log_location.start(LOG_LOCATION.MOD)
        self._curses_logwin.set_location(self._log_locations[self.log_location.value])

        # if to show kd ratio in User.moniker.
        self.show_kd = Toggle("_kd", [False, True])

        # how to sort users in the scoreboard.
        self.sort_order = Toggle("_so", SORT_ORDER)
        self.sort_order.start(SORT_ORDER.KD)
        self.set_sort_order(self.sort_order.value)

        #
        self.notify_operator = False
        self.sound_alarm = False

    def _build_grid(self):
        """Add boxes to the grid.

        Called during init and on KEY_RESIZE events.

        +-------------------+----------------------+
        | chatwin_blu       | chatwin_red          |  section 1
        |                   |                      |
        +-------------------+----------------------+
        | scorewin_blu      | scorewin_red         |  section 2
        |                   |                      |
        +--------------+----+----------------------+
        | user_win     | logger_win                |  section 3
        |              |                           |
        +--------------+--------+------------------+
        | status_win            | cmdline_win      |  section 4
        +-----------------------+------------------+
        """

        try:
            self._do_build_grid()
        except AssertionError:
            curses.endwin()
            logger.error("Terminal too small; try `Maximize` and `Ctrl+Minus`.")
            sys.exit(0)

    def _do_build_grid(self):

        # section 1
        self.chatwin_blu = self.grid.box(
            "chatwin_blu",
            nlines=10,
            ncols=int(self.grid.ncols / 2),
            left=self.grid,
            top=self.grid,
        )

        self.chatwin_red = self.grid.box(
            "chatwin_red",
            nlines=0,
            ncols=0,
            left2r=self.chatwin_blu,
            right=self.grid,
            top=self.chatwin_blu,
            bottom=self.chatwin_blu,
        )

        # section 2
        self.scorewin_blu = self.grid.box(
            "scorewin_blu",
            nlines=int(self.max_users / 2) + 2 + 1,  # 2=borders (top and bottom), 1=header.
            ncols=int(self.grid.ncols / 2),
            left=self.grid,
            top2b=self.chatwin_blu,
        )

        self.scorewin_red = self.grid.box(
            "scorewin_red",
            nlines=0,
            ncols=0,
            left2r=self.scorewin_blu,
            right=self.grid,
            top=self.scorewin_blu,
            bottom=self.scorewin_blu,
        )

        # section 4
        self.status_win = self.grid.box(
            "status",
            nlines=3,
            ncols=int(2 * self.grid.ncols / 3),
            left=self.grid,
            bottom=self.grid,
        )

        self.cmdline_win = self.grid.box(
            "cmdline",
            nlines=3,
            ncols=0,
            left2r=self.status_win,
            right=self.grid,
            bottom=self.grid,
        )

        # section 3 - fills gap between sections 2 and 4
        self.user_win = self.grid.box(
            "user",
            nlines=0,
            ncols=int(self.grid.ncols / 3),
            left=self.grid,
            top2b=self.scorewin_blu,
            bottom2t=self.status_win,
        )

        self.logger_win = self.grid.box(
            "logwin",
            nlines=0,
            ncols=0,
            left2r=self.user_win,
            right=self.grid,
            top=self.user_win,
            bottom=self.user_win,
        )

        #
        self.chatwin_blu.scrollok(True)
        self.chatwin_red.scrollok(True)
        self.scorewin_blu.scrollok(False)
        self.scorewin_red.scrollok(False)
        self.user_win.scrollok(True)
        self.logger_win.scrollok(True)
        self.status_win.scrollok(False)
        self.cmdline_win.scrollok(True)
        #
        self.cmdline_win.keypad(True)
        #
        self.redraw()

    def cycle_log_location(self):
        """Cycle format of location in messages displayed in logger window."""

        self._curses_logwin.set_location(self._log_locations[self.log_location.cycle])

    def set_sort_order(self, sort_order):
        """Set scoreboard sort column."""

        self.monitor.users.set_sort_order(sort_order)
        self._scoreboard.set_sort_order(sort_order)

    def getline(self, prompt=None):
        """Read and return next line from keyboard."""

        self.cmdline_win.erase()

        if prompt:
            self.cmdline_win.addstr(0, 0, prompt)
            self.cmdline_win.noutrefresh()

        return libcurses.getline(self.cmdline_win)

    def update_display(self):
        """Update display."""

        if self.sound_alarm:
            self.sound_alarm = False
            if self.monitor.conlog.is_eof:  # don't do this when replaying logfile from start
                # playsound('/usr/share/sounds/sound-icons/prompt.wav')
                # playsound('/usr/share/sounds/sound-icons/cembalo-10.wav')
                curses.flash()
                # curses.beep()

        #
        if self.monitor.kicks.msgs:
            # commandeer user window
            self._show_lines("KICKS", reversed(self.monitor.kicks.msgs), self.user_win)

        elif self.user_panel.value == USER_PANEL.SPAMS:
            self._show_lines("SPAMS", reversed(self.monitor.spams.msgs), self.user_win)

        elif self.user_panel.value == USER_PANEL.USER:
            self._show_user(self.monitor.me)

        else:
            raise ValueError(f"user_panel {self.user_panel!r}")

        # chatwin_blu and chatwin_red are rendered from gameplay/_playerchat

        #
        self._scoreboard.show_scores(
            team1=list(self.monitor.users.active_team_users(Team.BLU)),
            team2=list(self.monitor.users.active_team_users(Team.RED)),
        )

        #
        self.show_status()

        #
        self.refresh()

    def user_color(self, user, color):
        """Return `color` to display `user` in scoreboard."""

        if user.display_level:
            color = self.colormap[user.display_level]

        if user == self.monitor.my.last_killer:
            color |= curses.A_BOLD | curses.A_ITALIC

        if user == self.monitor.my.last_victim:
            color |= curses.A_BOLD

        if user.selected:
            color |= curses.A_REVERSE

        if not user.team:
            color |= curses.A_UNDERLINE

        if user.cloner:
            color |= curses.A_BLINK

        return color

    def show_chat(self, chat):
        """Display (append) `chat` in appropriate team window."""

        line = f"{chat.seqno}: {chat.user.username:20.20}: {chat.msg}"

        win = self.chatwin_blu if chat.user.team == Team.BLU else self.chatwin_red
        color = self.colormap[chat.user.team.name if chat.user.team else "user"]
        if chat.teamflag:
            color |= curses.A_UNDERLINE

        if sum(win.getyx()):
            win.addch("\n")
        win.addstr(line, self.user_color(chat.user, color))
        win.noutrefresh()

    def clear_chats(self):
        """Clear the chat windows."""

        now = time.asctime()
        for win in (self.chatwin_blu, self.chatwin_red):
            win.erase()
            win.addstr(now)

    #    def _show_chats(self, chats):
    #
    #        blu, red = [], []
    #
    #        for chat in chats:
    #            s = f'{chat.seqno}: {chat.user.username:.20}: {chat.msg}\n'
    #            if chat.team == Team.BLU:
    #                blu.append(s)
    #            else:
    #                red.append(s)
    #
    #        self._show_lines('BLU', blu, self.chatwin_blu)
    #        self._show_lines('RED', red, self.chatwin_red)

    def _show_user(self, user):

        lines = []
        indent = " " * 12  # 12=len("99 and 99 vs")

        lines.append("Duels:")
        for opponent in [x for x in user.opponents if x.state == UserState.ACTIVE]:
            lines.append(f"{user.duel_as_str(opponent, True)} vs {opponent.moniker}")

            if kills := user.kills.get(opponent):
                for weapon, count in kills.items():
                    lines.append(f"{indent} K {count:2} {weapon}")

            if deaths := user.deaths.get(opponent):
                for weapon, count in deaths.items():
                    lines.append(f"{indent} D {count:2} {weapon}")

        if self.show_actions.value:
            lines.extend(user.actions)

        if self.show_chats.value:
            lines.extend([x.msg for x in user.chats])

        self._show_lines("user", lines, self.user_win)

    def show_status(self):
        """Update status line."""

        line = self.monitor.fkeys.get_status_line() + f" UID={self.monitor.my.userid}"

        try:
            self.status_win.addstr(
                0, 0, line, self.colormap["NOTIFY" if self.notify_operator else "STATUS"]
            )
            self.status_win.clrtoeol()
        except curses.error:
            pass
        self.status_win.noutrefresh()

    def _show_lines(self, level, lines, win):

        win.erase()

        for line in lines:
            with contextlib.suppress(curses.error):
                win.addstr(line + "\n", self.colormap[level])

    def show_help(self):
        """Show help."""

        motd = self.monitor.tf2_cfg_dir / "motd.txt"
        logger.log("help", f" {motd} ".center(80, "-"))

        with open(motd, encoding="utf-8") as file:
            for line in file:
                logger.log("help", line.strip())
